#pragma pack(4)
#define WIN32_LEAN_AND_MEAN

#include <windows.h>

#include <stdio.h>
#include <conio.h>

#include <stdlib.h>
#include <winsock.h>


sockaddr_in CLIENT_GAME;
sockaddr_in CLIENT_PING;


sockaddr_in SERVER_GAME;
sockaddr_in SERVER_PING;


void WriteLog(unsigned char * Packet, int Length)
{
	printf("S2C 0000  ");

	for(int cnt = 0; cnt < Length; cnt++)
	{
		if(cnt && cnt % 16 == 0)
			printf("\nS2C %04X  ", cnt);
		printf("%02X ", Packet[cnt]);
	}
	printf("\n\n");
}


void ReadLog(unsigned char * Packet, int Length)
{
	printf("C2S 0000  ");

	for(int cnt = 0; cnt < Length; cnt++)
	{
		if(cnt && cnt % 16 == 0)
			printf("\nC2S %04X  ", cnt);
		printf("%02X ", Packet[cnt]);
	}
	printf("\n\n");
}


class Mutex
{
	HANDLE hMutex;

public:
	Mutex()
	{
		hMutex = CreateMutex(NULL, false, NULL);
	}

	virtual ~Mutex()
	{
		CloseHandle(hMutex);
		hMutex = NULL;
	}

	void LockClass() const
	{
		WaitForSingleObject(hMutex, INFINITE);
	}

	void UnlockClass() const
	{
		ReleaseMutex(hMutex);
	}

	inline Mutex(const Mutex &copy)
	{
		hMutex = copy.hMutex;
	}
};


class WinsockSession : public Mutex
{
	/* Threads */
	HANDLE hReceive, hPing;	// Handle to the receive threads

	void RecvThread()
	{
		// Receive queued messages
		sockaddr_in SockAddy;
		int FromLen = 16;
		BYTE buffer[520];

		while (hReceive)
		{
			int datalen = recvfrom(SocketID1, (char*)buffer, 520, 0, (sockaddr*)&SockAddy, &FromLen);

			if (datalen > 0)
			{
LockClass();
				if ((SockAddy.sin_addr.S_un.S_addr == SERVER_GAME.sin_addr.S_un.S_addr) && (SockAddy.sin_port == SERVER_GAME.sin_port))
				{
					// S2C packet
					sendto(SocketID1, (char*)&buffer[0], datalen, 0, (sockaddr*)&CLIENT_GAME, 16);
//					printf("S2C Packet\n");
					WriteLog(buffer, datalen);
				}
				else
				{
					// C2S packet
					sendto(SocketID1, (char*)&buffer[0], datalen, 0, (sockaddr*)&SERVER_GAME, 16);
//					printf("C2S Packet\n");
					CLIENT_GAME = SockAddy;
					ReadLog(buffer, datalen);
				}
UnlockClass();
			}
		}
	}

	friend DWORD WINAPI CallRecvThread(WinsockSession * ThisInstance)
	{
		ThisInstance->RecvThread();

		return 0;
	}

	void PingThread()
	{
		// Receive queued messages
		sockaddr_in SockAddy;
		int FromLen = 16;
		BYTE buffer[520];

		while (hPing)
		{
			int datalen = recvfrom(SocketID2, (char*)buffer, 520, 0, (sockaddr*)&SockAddy, &FromLen);

			if (datalen > 0)
			{
LockClass();
				if ((SockAddy.sin_addr.S_un.S_addr == SERVER_PING.sin_addr.S_un.S_addr) && (SockAddy.sin_port == SERVER_PING.sin_port))
				{
					// S2C packet
					sendto(SocketID2, (char*)&buffer[0], datalen, 0, (sockaddr*)&CLIENT_PING, 16);
//					printf("S2C Packet\n");
				}
				else
				{
					// C2S packet
					sendto(SocketID2, (char*)&buffer[0], datalen, 0, (sockaddr*)&SERVER_PING, 16);
//					printf("C2S Packet\n");
					CLIENT_PING = SockAddy;
				}
UnlockClass();
			}
		}
	}

	friend DWORD WINAPI CallPingThread(WinsockSession * ThisInstance)
	{
		ThisInstance->PingThread();

		return 0;
	}

public:
	/* Description */
	SOCKET	SocketID1;	// Session with winsock1 game
	SOCKET	SocketID2;	// Session with winsock2 ping
	SHORT	LocalPort1;	// Local port1 game
	SHORT	LocalPort2;	// Local port2 ping

	/* Initialization */
	void Initialize(SHORT NewPort)
	{
LockClass();
		// Kill old session
		if (hReceive)
			Terminate();

		// Create session with Winsock
		WSADATA wsaData;
		WSAStartup(0x202, &wsaData);

		// Register a socket
		LocalPort1 = NewPort;
		LocalPort2 = NewPort + 1;

		sockaddr_in sa;
					sa.sin_addr.S_un.S_addr	= 0;
					sa.sin_family			= 2;

					sa.sin_port				= htons(LocalPort1);

		SocketID1 = socket(2, 2, 0);
		bind(SocketID1, (sockaddr *) &sa, sizeof(sockaddr));

					sa.sin_port				= htons(LocalPort2);

		SocketID2 = socket(2, 2, 0);
		bind(SocketID2, (sockaddr *) &sa, sizeof(sockaddr));

		// Start threads
		unsigned long iThread;

		hReceive	= CreateThread(NULL, 0, (unsigned long (__stdcall *)(void *))&CallRecvThread,	this, 0, &iThread);
		hPing		= CreateThread(NULL, 0, (unsigned long (__stdcall *)(void *))&CallPingThread,	this, 0, &iThread);
UnlockClass();
	}

	void Terminate()
	{
LockClass();
		// Kill threads
		unsigned long ExitCode;

		GetExitCodeThread(hReceive, &ExitCode);
		TerminateThread(hReceive, ExitCode);

		GetExitCodeThread(hPing, &ExitCode);
		TerminateThread(hPing, ExitCode);

		// Unregister threads
		CloseHandle(hReceive);
		CloseHandle(hPing);

		// Unregister socket
		closesocket(SocketID1);
		closesocket(SocketID2);

		// Destroy session with Winsock
		WSACleanup();

		hReceive = 0;
UnlockClass();
	}

	WinsockSession() : Mutex()
	{
		hReceive = 0;
	}

	~WinsockSession()
	{
		if (hReceive)
			Terminate();
	}
};


LONG ResolveHostname(char * Hostname)
{
	LONG RetIP = inet_addr(Hostname);

	if (RetIP == -1)
	{
		hostent * host = gethostbyname(Hostname);

		if (host)
			RetIP = *((LONG*)*(host->h_addr_list));
		else
			RetIP = -1;
	}

	return RetIP;
}


void main()
{
	char buffer[64];

	// Load settings -> Set up server mapping
	char Path[520];
	GetCurrentDirectory(520, Path);

	// Log file
	char Logfile[520];
	memset(buffer, 0, 64);
	GetPrivateProfileString("Proxy", "Log", "packets.txt", buffer, 60, Path);
	strcpy(Logfile, Path);
	strcat(Logfile, "\\");
	strcat(Logfile, buffer);

	strcat(Path, "\\SSPROXY.INI");

	memset(buffer, 0, 32);
	GetPrivateProfileString("Proxy", "Port", "15000", buffer, 16, Path);
	SHORT Port = atoi(buffer);

	// Start winsock
	WinsockSession ws;
	ws.Initialize(Port);

	memset(buffer, 0, 32);
	GetPrivateProfileString("Server", "Port", "2000", buffer, 16, Path);
		SERVER_GAME.sin_port				= htons(atoi(buffer));
		SERVER_PING.sin_port				= htons(atoi(buffer) + 1);
	printf("Server Port = %i\n", atoi(buffer));

	memset(buffer, 0, 64);
	GetPrivateProfileString("Server", "IP", "127.0.0.1", buffer, 60, Path);
		SERVER_GAME.sin_addr.S_un.S_addr	= ResolveHostname(buffer);
		SERVER_PING.sin_addr.S_un.S_addr	= SERVER_GAME.sin_addr.S_un.S_addr;

		LONG IP = SERVER_GAME.sin_addr.S_un.S_addr;
	printf("Server IP = %i.%i.%i.%i\n", (int)((unsigned char*)&IP)[0]
									  , (int)((unsigned char*)&IP)[1]
									  , (int)((unsigned char*)&IP)[2]
									  , (int)((unsigned char*)&IP)[3]);

		SERVER_GAME.sin_family				= 2;
		SERVER_PING.sin_family				= 2;

	printf("SSProxy BUILD 5 - Ready to rock on port %i!\n", Port);
	printf("NOTE: Press any key to terminate.");

	freopen(Logfile, "wt", stdout);

	// Close on kbhit
	while (!kbhit())
	{
		Sleep(500);
	};

	// Bye!
	ws.Terminate();
}
